import sys
import os
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', 'a3cosmos_gas_evolution', 'Common_Python_Code'))

import numpy as np
import pytest
from calc_star_formation_law import (
    calc_Mmolgas_from_SFR_using_SF_law_Sargent2014,
    calc_SFR_from_Mmolgas_using_SF_law_Sargent2014,
)


class TestCalcStarFormationLaw:

    def test_calc_Mmolgas_sargent2014_scalar(self):
        SFR = 10.0
        DeltaMS = 0.0
        result = calc_Mmolgas_from_SFR_using_SF_law_Sargent2014(SFR, DeltaMS)
        assert np.isscalar(result) or result.shape == ()
        assert result > 0

    def test_calc_Mmolgas_sargent2014_array_SFR(self):
        SFR = np.array([1.0, 10.0, 100.0])
        DeltaMS = 0.0
        result = calc_Mmolgas_from_SFR_using_SF_law_Sargent2014(SFR, DeltaMS)
        assert len(result) == 3
        assert np.all(result > 0)

    def test_calc_Mmolgas_sargent2014_DeltaMS_low(self):
        SFR = 10.0
        DeltaMS = 0.0
        result = calc_Mmolgas_from_SFR_using_SF_law_Sargent2014(SFR, DeltaMS)
        assert result > 0

    def test_calc_Mmolgas_sargent2014_DeltaMS_high(self):
        SFR = 10.0
        DeltaMS = 1.0
        result = calc_Mmolgas_from_SFR_using_SF_law_Sargent2014(SFR, DeltaMS)
        assert result > 0

    def test_calc_Mmolgas_sargent2014_DeltaMS_array(self):
        SFR = 10.0
        DeltaMS = np.array([0.0, 0.5, 1.0])
        result = calc_Mmolgas_from_SFR_using_SF_law_Sargent2014(SFR, DeltaMS)
        assert len(result) == 3
        assert np.all(result > 0)

    def test_calc_SFR_from_Mmolgas_sargent2014_scalar(self):
        Mmolgas = 1e10
        DeltaMS = 0.0
        result = calc_SFR_from_Mmolgas_using_SF_law_Sargent2014(Mmolgas, DeltaMS)
        assert np.isscalar(result) or result.shape == ()
        assert result > 0

    def test_calc_SFR_from_Mmolgas_sargent2014_array(self):
        Mmolgas = np.array([1e9, 1e10, 1e11])
        DeltaMS = 0.0
        result = calc_SFR_from_Mmolgas_using_SF_law_Sargent2014(Mmolgas, DeltaMS)
        assert len(result) == 3
        assert np.all(result > 0)

    def test_roundtrip_consistency(self):
        SFR_original = 10.0
        DeltaMS = 0.5
        Mmol = calc_Mmolgas_from_SFR_using_SF_law_Sargent2014(SFR_original, DeltaMS)
        SFR_recovered = calc_SFR_from_Mmolgas_using_SF_law_Sargent2014(Mmol, DeltaMS)
        assert np.isclose(SFR_original, SFR_recovered, rtol=1e-5)

    def test_Mmolgas_increases_with_SFR(self):
        SFR_low = 1.0
        SFR_high = 100.0
        DeltaMS = 0.0
        result_low = calc_Mmolgas_from_SFR_using_SF_law_Sargent2014(SFR_low, DeltaMS)
        result_high = calc_Mmolgas_from_SFR_using_SF_law_Sargent2014(SFR_high, DeltaMS)
        assert result_high > result_low

    def test_Mmolgas_increases_with_DeltaMS(self):
        SFR = 10.0
        DeltaMS_low = 0.0
        DeltaMS_high = 1.0
        result_low = calc_Mmolgas_from_SFR_using_SF_law_Sargent2014(SFR, DeltaMS_low)
        result_high = calc_Mmolgas_from_SFR_using_SF_law_Sargent2014(SFR, DeltaMS_high)
        assert result_low > 0 and result_high > 0



#!/usr/bin/env python
# 
# 20190222
#     copied from "calc_stellar_mass_function.py", this code will superceed "calc_stellar_mass_function.py". 
# 

from __future__ import print_function

import os, sys, re, json, time, astropy
import numpy as np
from astropy.table import Table, Column, hstack
from copy import copy
from numpy import log, log10, power, sum, sqrt, pi, exp
pow = power
lg = log10
ln = log
from scipy.interpolate import InterpolatedUnivariateSpline, interp1d

if not (os.path.dirname(os.path.abspath(__file__)) in sys.path): sys.path.append(os.path.dirname(os.path.abspath(__file__)))
import apply_cosmology
cosmo = apply_cosmology.cosmo

if sys.version_info.major >= 3:
    long = int
else:
    pass




# 
# def 
# 
def Schechter_Function_for_LF(L, L_character, Phi_character, alpha):
    # 
    # Schechter (1976)
    # 
    # Phi(L) dL = (Phi_*) * (L/L_*)**(alpha) * exp(-L/L_*) dL/L_*
    #           = (Phi_*) * x**(alpha) * exp(-x) dx
    #           = (Phi_*) * 10**(lgx * alpha) * exp(-10**lgx) dx
    #           = (Phi_*) * 10**(lgx * alpha) * exp(-10**lgx) dlnx
    #           = (Phi_*) * 10**(lgx * alpha) * exp(-10**lgx) dlgx * ln(10)
    #           = (Phi_*) * 10**((lgL-lgL_*)*(alpha+1)) * exp(-10**(lgL-lgL_*)) * ln(10)  dlgx
    #           = (Our_Phi_Phi_Schechter)                                                 dlgx
    # 
    #lgx = lgL-lg_L0
    #Phi_Schechter = phi * (10**(lgx*(alpha+1))) * (np.exp(-10**lgx)) * ln(10) # per dex and already multiplied ln(10), so that its integral directly equals \int Phi(L) / L dL
    # 
    Phi_Schechter = Phi_character * (L/L_character)**(alpha) * np.exp(-(L/L_character)) # Mpc-3 dex-1
    #Phi_Schechter = Phi_Schechter * ln(10)
    return Phi_Schechter


def Saunders_Function_for_LF(L, L_character, Phi_character, alpha, sigma):
    # Saunders et al. (1990)
    Phi_Saunders = Phi_character * (L/L_character)**(1-alpha) * np.exp(-1.0/(2.0*sigma**2) * (np.log10(1.0+(L/L_character)))**2 )
    #print('Phi_character', Phi_character)
    #print('(L/L_character)**(1-alpha)', (L/L_character)**(1-alpha))
    #print('np.exp(-1.0/(2.0*sigma**2) * (np.log10(1.0+(L/L_character)))**2 )', np.exp(-1.0/(2.0*sigma**2) * (np.log10(1.0+(L/L_character)))**2 ))
    #print('Phi_Saunders', Phi_Saunders)
    return Phi_Saunders





# 
# def 
# 
def calc_radio_LF_Novak2017(z, lgL=None, galaxy_type = 'SFG'):
    # 
    # Novak 2017 bibcode:2017A&A...602A...5N
    # IMF: Chabrier 2003
    # Saunders et al. (1990)
    # Outputs: lgL_grid, lgPhi_grid
    # 
    # check z
    if not np.isscalar(z):
        raise ValueError('Please input a float number as the redshift!')
    # 
    # check galaxy_type
    if not (type(galaxy_type) is str):
        raise ValueError('Please input either "ALL", "SFG" or "QG" as the galaxy_type!')
    else:
        if not (galaxy_type in ['ALL', 'SFG', 'QG']):
            raise ValueError('Please input either "ALL", "SFG" or "QG" as the galaxy_type!')
    # 
    # make lgL_grid
    if lgL is None:
        lgL_grid = np.linspace(18.0, 25.0, num=1000, endpoint=True)
    else:
        lgL_grid = lgL
    # 
    L_grid = 10**lgL_grid
    # 
    # read LF parameters
    L_character = 1.85e21 # * 1.4e9 / 3.839e25 # vLv(1.4GHz,rest) = W Hz-1 --> Lsun
    Phi_character = 3.55e-3 # Mpc-3 dex-1
    alpha = 1.22
    sigma = 0.63
    # 
    #Phi_z0 = Saunders_Function(L_grid, L_character, Phi_character, alpha, sigma)
    # 
    # check z
    LF_zmin = 0.0
    LF_zmax = +np.inf
    if z < LF_zmin or z > LF_zmax:
        raise ValueError('calc_radio_LF_Novak2017: The input redshift is out of the allowed range of %s -- %s!'%(LF_zmin, LF_zmax))
    # 
    # scale to z via pure luminosity evolution
    alphaL = 3.16
    betaL = -0.32
    L_grid_z = (L_grid / ((1.0+z)**(alphaL+(z*betaL))))
    Phi = Saunders_Function_for_LF(L_grid_z, L_character, Phi_character, alpha, sigma)
    lgPhi = np.log10(Phi)
    # 
    if lgL is None:
        return lgL_grid, lgPhi
    else:
        return lgPhi



def calc_IR_250um_LF_Koprowski2017(z, lgL=None, galaxy_type = 'SFG'):
    # 
    # Koprowski 2017 bibcode:2017MNRAS.471.4155K
    # IMF: Chabrier 2003
    # Saunders et al. (1990)
    # Outputs: lgL_grid, lgPhi_grid
    # 
    # check z
    if not np.isscalar(z):
        raise ValueError('Please input a float number as the redshift!')
    # 
    # check galaxy_type
    if not (type(galaxy_type) is str):
        raise ValueError('Please input either "ALL", "SFG" or "QG" as the galaxy_type!')
    else:
        if not (galaxy_type in ['ALL', 'SFG', 'QG']):
            raise ValueError('Please input either "ALL", "SFG" or "QG" as the galaxy_type!')
    # 
    # make lgL_grid
    if lgL is None:
        lgL_grid = np.linspace(24.0, 27.0, num=1000, endpoint=True)
    else:
        lgL_grid = lgL
    # 
    L_grid = 10**lgL_grid
    # 
    # read LF parameters
    table_z_lower = [0.5, 1.5, 2.5, 3.5]
    table_z_upper = [1.5, 2.5, 3.5, 4.5]
    table_lgL_character = [25.20, 25.40, 25.63, 25.84] # W Hz-1
    table_lgPhi_character = [-2.88, -3.03, -3.73, -4.59] # Mpc-3 dex-1
    alpha = -0.4
    # 
    # check z
    LF_zmin = table_z_lower[0]
    LF_zmax = table_z_upper[-1]
    if z < LF_zmin or z > LF_zmax:
        raise ValueError('calc_IR_250um_LF_Koprowski2017: The input redshift is out of the allowed range of %s -- %s!'%(LF_zmin, LF_zmax))
    # 
    # scale to z (using step function... <TODO>)
    Phi = None
    lgPhi = None
    for i in range(len(table_z_upper)):
        if z >= table_z_lower[i] and z <= table_z_upper[i]:
            L_character = 10**(table_lgL_character[i])
            Phi_character = 10**(table_lgPhi_character[i])
            Phi = Schechter_Function_for_LF(L_grid, L_character, Phi_character, alpha)
            lgPhi = np.log10(Phi)
            break
    # 
    if lgL is None:
        return lgL_grid, lgPhi
    else:
        return lgPhi



def calc_IR_LF_Gruppioni2013(z, lgL=None, galaxy_type = 'SFG'):
    # 
    # Gruppioni 2013 bibcode:
    # IMF: Chabrier 2003
    # H0 = 71 km s−1 Mpc−1, Ωm = 0.27, and ΩΛ = 0.73.
    # Saunders et al. (1990)
    # Outputs: lgL_grid, lgPhi_grid
    # 
    # check z
    if not np.isscalar(z):
        raise ValueError('Please input a float number as the redshift!')
    # 
    # check galaxy_type
    if not (type(galaxy_type) is str):
        raise ValueError('Please input either "ALL", "SFG" or "QG" as the galaxy_type!')
    else:
        if not (galaxy_type in ['ALL', 'SFG', 'QG']):
            raise ValueError('Please input either "ALL", "SFG" or "QG" as the galaxy_type!')
    # 
    # make lgL_grid
    if lgL is None:
        lgL_grid = np.linspace(8.0, 14.0, num=1000, endpoint=True)
    else:
        lgL_grid = lgL
    # 
    L_grid = 10**lgL_grid
    # 
    # read LF parameters (their Table 7)
    table_data = [ [0.0 , 0.3 , 1.15, 0.52, 10.12, -2.29], 
                   [0.3 , 0.45, 1.2 , 0.5 , 10.41, -2.31], 
                   [0.45, 0.6 , 1.2 , 0.5 , 10.55, -2.35], 
                   [0.6 , 0.8 , 1.2 , 0.5 , 10.71, -2.35], 
                   [0.8 , 1.0 , 1.2 , 0.5 , 10.97, -2.40], 
                   [1.0 , 1.2 , 1.2 , 0.5 , 11.13, -2.43], 
                   [1.2 , 1.7 , 1.2 , 0.5 , 11.37, -2.70], 
                   [1.7 , 2.0 , 1.2 , 0.5 , 11.50, -3.00], 
                   [2.0 , 2.5 , 1.2 , 0.5 , 11.60, -3.01], 
                   [2.5 , 3.0 , 1.2 , 0.5 , 11.92, -3.27], 
                   [3.0 , 4.2 , 1.2 , 0.5 , 11.90, -3.74] ]
    table_data = np.array(table_data).T
    table_z_lower = table_data[0]
    table_z_upper = table_data[1]
    table_alpha = table_data[2]
    table_sigma = table_data[3]
    table_lgL_character = table_data[4] # Lsun
    table_lgPhi_character = table_data[5] # Mpc-3 dex-1
    # 
    # check z
    LF_zmin = table_z_lower[0]
    LF_zmax = table_z_upper[-1]
    if z < LF_zmin or z > LF_zmax:
        raise ValueError('calc_IR_LF_Gruppioni2013: The input redshift is out of the allowed range of %s -- %s!'%(LF_zmin, LF_zmax))
    # 
    # scale to z (using step function... <TODO>)
    Phi = None
    lgPhi = None
    for i in range(len(table_z_upper)):
        if z >= table_z_lower[i] and z <= table_z_upper[i]:
            L_character = 10**(table_lgL_character[i])
            Phi_character = 10**(table_lgPhi_character[i])
            alpha = table_alpha[i]
            sigma = table_sigma[i]
            Phi = Saunders_Function_for_LF(L_grid, L_character, Phi_character, alpha, sigma)
            lgPhi = np.log10(Phi)
            break
    # 
    if lgL is None:
        return lgL_grid, lgPhi
    else:
        return lgPhi


























import sys
import os
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', 'a3cosmos_gas_evolution', 'Common_Python_Code'))

import numpy as np
import pytest
from calc_metal_Z import (
    calc_metalZ_from_FMR_following_Kewley2008_PP04_O3N2,
    calc_metalZ_from_FMR_following_Kewley2008_PP04_N2,
    calc_metalZ_from_FMR_following_Kewley2008_KK04,
    calc_metalZ_from_FMR_following_Kewley2008_KD02,
    calc_metalZ_from_FMR_following_Mannucci2010,
    calc_metalZ_from_FMR_following_Mannucci2010_Eq2,
    calc_metalZ_from_FMR_following_Mannucci2010_Eq4,
    calc_metalZ_from_FMR_following_Mannucci2011,
    calc_metalZ_from_FMR_following_Maiolino2008,
    calc_metalZ_from_FMR_following_Genzel2015_Eq12a,
    calc_metalZ_from_FMR_following_Genzel2015_Eq12a_with_dzliu_limit,
    calc_metalZ_from_FMR_following_Genzel2015a,
    calc_metalZ_from_FMR_following_Genzel2015b,
    calc_metalZ_from_FMR_following_Genzel2015ab_combined_by_dzliu,
    calc_metalZ_from_FMR_following_Guo2016,
    convert_metalZ_M08_to_metalZ_PP04,
    convert_metalZ_D02_to_metalZ_PP04,
    convert_metalZ_KK04_to_metalZ_PP04,
)


class TestCalcMetalZ:

    def test_kewley2008_PP04_O3N2_scalar(self):
        result = calc_metalZ_from_FMR_following_Kewley2008_PP04_O3N2(1e10)
        assert np.isscalar(result) or result.shape == ()
        assert 7.0 < result < 10.0

    def test_kewley2008_PP04_O3N2_array(self):
        M_star = np.array([1e9, 1e10, 1e11])
        result = calc_metalZ_from_FMR_following_Kewley2008_PP04_O3N2(M_star)
        assert len(result) == 3

    def test_kewley2008_PP04_N2_scalar(self):
        result = calc_metalZ_from_FMR_following_Kewley2008_PP04_N2(1e10)
        assert np.isscalar(result) or result.shape == ()
        assert 7.0 < result < 10.0

    def test_kewley2008_KK04_scalar(self):
        result = calc_metalZ_from_FMR_following_Kewley2008_KK04(1e10)
        assert np.isscalar(result) or result.shape == ()

    def test_kewley2008_KD02_scalar(self):
        result = calc_metalZ_from_FMR_following_Kewley2008_KD02(1e10)
        assert np.isscalar(result) or result.shape == ()

    def test_mannucci2010_scalar(self):
        result = calc_metalZ_from_FMR_following_Mannucci2010(1e10, 1.0)
        assert np.isscalar(result) or result.shape == ()
        assert 7.0 < result < 10.0

    def test_mannucci2010_array_Mstar(self):
        M_star = np.array([1e9, 1e10, 1e11])
        SFR = 1.0
        result = calc_metalZ_from_FMR_following_Mannucci2010(M_star, SFR)
        assert len(result) == 3

    def test_mannucci2010_array_SFR(self):
        M_star = 1e10
        SFR = np.array([0.1, 1.0, 10.0])
        result = calc_metalZ_from_FMR_following_Mannucci2010(M_star, SFR)
        assert len(result) == 3

    def test_mannucci2010_Eq2_vs_Eq4(self):
        M_star = 1e10
        SFR = 1.0
        result_Eq2 = calc_metalZ_from_FMR_following_Mannucci2010_Eq2(M_star, SFR)
        result_Eq4 = calc_metalZ_from_FMR_following_Mannucci2010_Eq4(M_star, SFR)
        assert np.isscalar(result_Eq2) or result_Eq2.shape == ()
        assert np.isscalar(result_Eq4) or result_Eq4.shape == ()

    def test_mannucci2011_scalar(self):
        result = calc_metalZ_from_FMR_following_Mannucci2011(1e10, 1.0)
        assert np.isscalar(result) or result.shape == ()

    def test_maiolino2008_z0(self):
        result = calc_metalZ_from_FMR_following_Maiolino2008(1e10, 0.0)
        assert np.isscalar(result) or result.shape == ()

    def test_maiolino2008_high_z(self):
        result = calc_metalZ_from_FMR_following_Maiolino2008(1e10, 3.0)
        assert np.isscalar(result) or result.shape == ()

    def test_genzel2015_Eq12a_scalar(self):
        result = calc_metalZ_from_FMR_following_Genzel2015_Eq12a(1e10, 1.0)
        assert np.isscalar(result) or result.shape == ()

    def test_genzel2015_Eq12a_array(self):
        M_star = np.array([1e9, 1e10, 1e11])
        z = 1.0
        result = calc_metalZ_from_FMR_following_Genzel2015_Eq12a(M_star, z)
        assert len(result) == 3

    def test_genzel2015_Eq12a_with_dzliu_limit(self):
        result = calc_metalZ_from_FMR_following_Genzel2015_Eq12a_with_dzliu_limit(1e10, 1.0)
        assert np.isscalar(result) or result.shape == ()

    def test_genzel2015a_scalar(self):
        result = calc_metalZ_from_FMR_following_Genzel2015a(1e10, 1.0, 1.0)
        assert np.isscalar(result) or result.shape == ()

    def test_genzel2015b_scalar(self):
        result = calc_metalZ_from_FMR_following_Genzel2015b(1e10, 1.0)
        assert np.isscalar(result) or result.shape == ()

    def test_genzel2015ab_combined_scalar(self):
        result = calc_metalZ_from_FMR_following_Genzel2015ab_combined_by_dzliu(1e10, 1.0, 1.0)
        assert np.isscalar(result) or result.shape == ()

    def test_genzel2015ab_combined_array(self):
        M_star = np.array([1e9, 1e10, 1e11])
        SFR = 1.0
        z = 1.0
        result = calc_metalZ_from_FMR_following_Genzel2015ab_combined_by_dzliu(M_star, SFR, z)
        assert len(result) == 3

    def test_guo2016_scalar(self):
        result = calc_metalZ_from_FMR_following_Guo2016(1e10, 1.0)
        assert np.isscalar(result) or result.shape == ()

    def test_convert_M08_to_PP04(self):
        metalZ_M08 = 8.67
        result = convert_metalZ_M08_to_metalZ_PP04(metalZ_M08)
        assert np.isscalar(result) or result.shape == ()
        assert 7.0 < result < 10.0

    def test_convert_D02_to_PP04(self):
        metalZ_D02 = 8.67
        result = convert_metalZ_D02_to_metalZ_PP04(metalZ_D02)
        assert np.isscalar(result) or result.shape == ()

    def test_convert_KK04_to_PP04(self):
        metalZ_KK04 = 8.67
        result = convert_metalZ_KK04_to_metalZ_PP04(metalZ_KK04)
        assert np.isscalar(result) or result.shape == ()

    def test_metalZ_increases_with_stellar_mass(self):
        M_star_low = 1e9
        M_star_high = 1e11
        result_low = calc_metalZ_from_FMR_following_Kewley2008_PP04_O3N2(M_star_low)
        result_high = calc_metalZ_from_FMR_following_Kewley2008_PP04_O3N2(M_star_high)
        assert result_high > result_low



import sys
import os
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', 'a3cosmos_gas_evolution', 'Common_Python_Code'))

import numpy as np
import pytest

from calc_gas_mass_from_line import (
    ALPHA_CII,
    ALPHA_CO_MW,
    ALPHA_CO_SB,
    calc_gas_mass_from_line_flux,
)


class TestCalcGasMassFromLineConstants:

    def test_alpha_CII_value(self):
        assert ALPHA_CII == 30.0

    def test_alpha_CO_MW_value(self):
        assert ALPHA_CO_MW == 4.3

    def test_alpha_CO_SB_value(self):
        assert ALPHA_CO_SB == 0.8


class TestCalcGasMassFromLineBoundaryCases:

    def test_invalid_z_None_returns_nan(self):
        lgMgas, lgLline, line_unit, line_type = calc_gas_mass_from_line_flux(
            z=None, lgMstar=10.5, freq=115.27, flux=1.0, line_name='CO(1-0)'
        )
        assert np.isnan(lgMgas)

    def test_invalid_z_empty_string_returns_nan(self):
        lgMgas, lgLline, line_unit, line_type = calc_gas_mass_from_line_flux(
            z='', lgMstar=10.5, freq=115.27, flux=1.0, line_name='CO(1-0)'
        )
        assert np.isnan(lgMgas)

    def test_invalid_z_negative_returns_nan(self):
        lgMgas, lgLline, line_unit, line_type = calc_gas_mass_from_line_flux(
            z=-1.0, lgMstar=10.5, freq=115.27, flux=1.0, line_name='CO(1-0)'
        )
        assert np.isnan(lgMgas)

    def test_invalid_z_zero_returns_nan(self):
        lgMgas, lgLline, line_unit, line_type = calc_gas_mass_from_line_flux(
            z=0.0, lgMstar=10.5, freq=115.27, flux=1.0, line_name='CO(1-0)'
        )
        assert np.isnan(lgMgas)

    def test_invalid_z_nan_returns_nan(self):
        lgMgas, lgLline, line_unit, line_type = calc_gas_mass_from_line_flux(
            z=np.nan, lgMstar=10.5, freq=115.27, flux=1.0, line_name='CO(1-0)'
        )
        assert np.isnan(lgMgas)


class TestCalcGasMassFromLineCO:

    def test_co10_line_z2(self):
        lgMgas, lgLline, line_unit, line_type = calc_gas_mass_from_line_flux(
            z=2.0, lgMstar=10.5, freq=115.27, flux=1.0, line_name='CO(1-0)'
        )
        assert not np.isnan(lgMgas)
        assert not np.isnan(lgLline)
        assert line_type == 'CO'
        assert line_unit == 'K km s-1 pc2'
        assert lgMgas > 0
        assert lgLline > 0

    def test_co10_line_z1(self):
        lgMgas, lgLline, line_unit, line_type = calc_gas_mass_from_line_flux(
            z=1.0, lgMstar=10.5, freq=115.27 / 2.0, flux=1.0, line_name='CO(1-0)'
        )
        assert not np.isnan(lgMgas)
        assert line_type == 'CO'

    def test_co21_line_z2(self):
        lgMgas, lgLline, line_unit, line_type = calc_gas_mass_from_line_flux(
            z=2.0, lgMstar=10.5, freq=230.54, flux=1.0, line_name='CO(2-1)'
        )
        assert not np.isnan(lgMgas)
        assert line_type == 'CO'

    def test_co32_line_z2(self):
        lgMgas, lgLline, line_unit, line_type = calc_gas_mass_from_line_flux(
            z=2.0, lgMstar=10.5, freq=345.796, flux=1.0, line_name='CO(3-2)'
        )
        assert not np.isnan(lgMgas)
        assert line_type == 'CO'

    def test_co_line_high_z(self):
        lgMgas, lgLline, line_unit, line_type = calc_gas_mass_from_line_flux(
            z=5.0, lgMstar=10.5, freq=115.27 / 6.0, flux=1.0, line_name='CO(1-0)'
        )
        assert not np.isnan(lgMgas)
        assert lgMgas > 0

    def test_co_line_with_custom_alpha_CO(self):
        lgMgas, lgLline, line_unit, line_type = calc_gas_mass_from_line_flux(
            z=2.0, lgMstar=10.5, freq=115.27, flux=1.0, line_name='CO(1-0)',
            alpha_CO=3.0
        )
        assert not np.isnan(lgMgas)

    def test_co_line_without_logMstar(self):
        lgMgas, lgLline, line_unit, line_type = calc_gas_mass_from_line_flux(
            z=2.0, lgMstar=None, freq=115.27, flux=1.0, line_name='CO(1-0)'
        )
        assert not np.isnan(lgMgas)


class TestCalcGasMassFromLineCII:

    def test_cii_line_z6(self):
        lgMgas, lgLline, line_unit, line_type = calc_gas_mass_from_line_flux(
            z=6.0, lgMstar=10.5, freq=1900.0, flux=1.0, line_name='[CII]'
        )
        assert not np.isnan(lgMgas)
        assert not np.isnan(lgLline)
        assert line_type == 'CII'
        assert line_unit == 'L_sun'
        assert lgMgas > 0
        assert lgLline > 0

    def test_cii_line_alt_name(self):
        lgMgas, lgLline, line_unit, line_type = calc_gas_mass_from_line_flux(
            z=6.0, lgMstar=10.5, freq=1900.0, flux=1.0, line_name='CII'
        )
        assert not np.isnan(lgMgas)
        assert line_type == 'CII'

    def test_cii_line_high_z(self):
        lgMgas, lgLline, line_unit, line_type = calc_gas_mass_from_line_flux(
            z=7.0, lgMstar=10.5, freq=1900.0 / 8.0, flux=1.0, line_name='[CII]'
        )
        assert not np.isnan(lgMgas)
        assert lgMgas > 0

    def test_cii_line_with_custom_alpha_CII(self):
        lgMgas, lgLline, line_unit, line_type = calc_gas_mass_from_line_flux(
            z=6.0, lgMstar=10.5, freq=1900.0, flux=1.0, line_name='[CII]',
            alpha_CII=50.0
        )
        assert not np.isnan(lgMgas)


class TestCalcGasMassFromLineFlux:

    def test_different_flux_values(self):
        fluxes = [0.1, 1.0, 10.0]
        for flux in fluxes:
            lgMgas, lgLline, line_unit, line_type = calc_gas_mass_from_line_flux(
                z=2.0, lgMstar=10.5, freq=115.27, flux=flux, line_name='CO(1-0)'
            )
            assert not np.isnan(lgMgas)
            assert lgMgas > 0

    def test_flux_zero_returns_nan(self):
        lgMgas, lgLline, line_unit, line_type = calc_gas_mass_from_line_flux(
            z=2.0, lgMstar=10.5, freq=115.27, flux=0.0, line_name='CO(1-0)'
        )
        assert np.isnan(lgMgas)


class TestCalcGasMassFromLineOutput:

    def test_output_is_4_tuple(self):
        result = calc_gas_mass_from_line_flux(
            z=2.0, lgMstar=10.5, freq=115.27, flux=1.0, line_name='CO(1-0)'
        )
        assert len(result) == 4

    def test_co_and_cii_output_shapes_consistent(self):
        result_co = calc_gas_mass_from_line_flux(
            z=2.0, lgMstar=10.5, freq=115.27, flux=1.0, line_name='CO(1-0)'
        )
        result_cii = calc_gas_mass_from_line_flux(
            z=6.0, lgMstar=10.5, freq=1900.0, flux=1.0, line_name='[CII]'
        )
        assert len(result_co) == len(result_cii) == 4

    def test_output_shapes_consistent(self):
        result_co = calc_gas_mass_from_line_flux(
            z=2.0, lgMstar=10.5, freq=115.27, flux=1.0, line_name='CO(1-0)'
        )
        result_cii = calc_gas_mass_from_line_flux(
            z=6.0, lgMstar=10.5, freq=1900.0, flux=1.0, line_name='[CII]'
        )
        assert len(result_co) == len(result_cii) == 4

    def test_lgLline_increases_with_flux(self):
        lgLline_low = calc_gas_mass_from_line_flux(
            z=2.0, lgMstar=10.5, freq=115.27, flux=0.5, line_name='CO(1-0)'
        )[1]
        lgLline_high = calc_gas_mass_from_line_flux(
            z=2.0, lgMstar=10.5, freq=115.27, flux=1.0, line_name='CO(1-0)'
        )[1]
        assert lgLline_high > lgLline_low



#!/usr/bin/env python
# 
# Usage:
#    from calc_metal_Z import ( calc_metalZ_from_FMR_following_Mannucci2010, 
#                               calc_metalZ_from_FMR_following_Genzel2015a, 
#                               calc_metalZ_from_FMR_following_Genzel2015b, 
#                               calc_metalZ_from_FMR_following_Genzel2015ab_combined_by_dzliu, 
#                               convert_metalZ_M08_to_metalZ_PP04, 
#                               convert_metalZ_D02_to_metalZ_PP04, 
#                               convert_metalZ_KK04_to_metalZ_PP04, 
#                             )
# 
# 

from __future__ import print_function

import os, sys, re, json, time, astropy
import numpy as np
from astropy.table import Table, Column, hstack
from copy import copy

if not (os.path.dirname(os.path.abspath(__file__)) in sys.path): sys.path.append(os.path.dirname(os.path.abspath(__file__)))
import apply_cosmology
cosmo = apply_cosmology.cosmo

if sys.version_info.major >= 3:
    long = int
else:
    pass



# 
# def 
# 
def calc_metalZ_from_FMR_following_Kewley2008_PP04_O3N2(M_star):
    # metalZ is 12+log(O/H)_{PP04(O3N2)}
    # 
    # Kewley & Ellison (2008ApJ...681.1183K)
    # Table 2
    # Note that their caption has typo: a + b*x + c*x**2 + d*x**3
    x = np.log10(M_star)
    a = 32.1488
    b = -8.51258
    c = 0.976384
    d = -0.0359763
    metalZ_PP04 = a + b*x + c*x**2 + d*x**3
    # 
    return metalZ_PP04


def calc_metalZ_from_FMR_following_Kewley2008_PP04_N2(M_star):
    # metalZ is 12+log(O/H)_{PP04(N2)}
    # 
    # Kewley & Ellison (2008ApJ...681.1183K)
    # Table 2
    # Note that their caption has typo: a + b*x + c*x**2 + d*x**3
    x = np.log10(M_star)
    a = 23.9049
    b = -5.62784
    c = 0.645142
    d = -0.0235065
    metalZ_PP04 = a + b*x + c*x**2 + d*x**3
    # 
    return metalZ_PP04


def calc_metalZ_from_FMR_following_Kewley2008_KK04(M_star):
    # metalZ is 12+log(O/H)_{KK04(R23)}
    # 
    # Kewley & Ellison (2008ApJ...681.1183K)
    # Table 2
    # Note that their caption has typo: a + b*x + c*x**2 + d*x**3
    x = np.log10(M_star)
    a = 27.7911
    b = -6.94493
    c = 0.808097
    d = -0.0301508
    metalZ_KK04 = a + b*x + c*x**2 + d*x**3
    # 
    return metalZ_KK04


def calc_metalZ_from_FMR_following_Kewley2008_KD02(M_star):
    # metalZ is 12+log(O/H)_{KD02}
    # 
    # Kewley & Ellison (2008ApJ...681.1183K)
    # Table 2
    # Note that their caption has typo: a + b*x + c*x**2 + d*x**3
    x = np.log10(M_star)
    a = 28.0974
    b = -7.23631
    c = 0.850344
    d = -0.0318315
    metalZ_KD02 = a + b*x + c*x**2 + d*x**3
    # 
    return metalZ_KD02


def calc_metalZ_from_FMR_following_Mannucci2010(M_star, SFR):
    # metalZ is 12+log(O/H)_{Maiolino2008}
    # 
    # Mannucci2010
    # Sect. 4, Eq. 2
    m = np.log10(M_star) - 10
    s = np.log10(SFR)
    metalZ_M08 = 8.90 + 0.37*m - 0.14*s - 0.19*m**2 + 0.12*m*s - 0.054*s**2
    # 
    return metalZ_M08


def calc_metalZ_from_FMR_following_Mannucci2010_Eq2(M_star, SFR):
    # metalZ is 12+log(O/H)_{Maiolino2008}
    # 
    # Mannucci2010
    # Sect. 4, Eq. 2, two variables
    m = np.log10(M_star) - 10
    s = np.log10(SFR)
    metalZ_M08 = 8.90 + 0.37*m - 0.14*s - 0.19*m**2 + 0.12*m*s - 0.054*s**2
    # 
    return metalZ_M08


def calc_metalZ_from_FMR_following_Mannucci2010_Eq4(M_star, SFR):
    # metalZ is 12+log(O/H)_{Maiolino2008}
    # 
    # Mannucci2010
    # Sect. 4, Eq. 4, a single variable
    x = np.log10(M_star) - 0.32 * np.log10(SFR) - 10
    metalZ_M08 = 8.90 + 0.39 * x - 0.20 * x**2 - 0.077 * x**3 + 0.064 * x**4
    # 
    return metalZ_M08


def calc_metalZ_from_FMR_following_Mannucci2011(M_star, SFR):
    # metalZ is 12+log(O/H)_{Maiolino2008}
    # 
    # Mannucci 2011 - 2011MNRAS.414.1263M.pdf
    # Sect. 2, Eq. 2
    m = np.log10(M_star) - 10
    s = np.log10(SFR)
    mu032 = np.log10(M_star) - 0.32 * np.log10(SFR)
    x = np.log10(M_star) - 0.32 * np.log10(SFR) - 10
    if np.isscalar(mu032):
        if mu032 >= 9.5:
            metalZ_M08 = 8.90 + 0.37*m - 0.14*s - 0.19*m**2 + 0.12*m*s - 0.054*s**2
            #metalZ_M08 = 8.90 + 0.39 * x - 0.20 * x**2 - 0.077 * x**3 + 0.064 * x**4
        else:
            metalZ_M08 = 8.93 + 0.51*(mu032-10)
    else:
        if np.isscalar(s) and ~np.isscalar(m):
            s = m*0.0 + s
        if np.isscalar(m) and ~np.isscalar(s):
            m = s*0.0 + m
        mask = (mu032 >= 9.5)
        metalZ_M08 = 8.90 + 0.37*m - 0.14*s - 0.19*m**2 + 0.12*m*s - 0.054*s**2
        #metalZ_M08 = 8.90 + 0.39 * x - 0.20 * x**2 - 0.077 * x**3 + 0.064 * x**4
        metalZ_M08[~mask] = 8.93 + 0.51*(mu032[~mask]-10)
    # 
    return metalZ_M08


def calc_metalZ_from_FMR_following_Maiolino2008(M_star, z):
    # metalZ is 12+log(O/H)_{Maiolino2008}
    # 
    # Maiolino2008 doi:10.1051/0004-6361/201220074
    # Sect. 7.6, Eq. 2
    # 12+log(O/H) = -0.0864 * (logMstar - logM0)**2 + K0
    # Table 5
    #   z     logM0  K0
    #   0.07  11.18  9.04
    #   0.7   11.57  9.04
    #   2.2   12.38  8.99
    #   3.5   12.76  8.79
    #   3.5   12.87  8.90
    # 
    m = np.log10(M_star)
    spl_z = np.array([0.07,0.7,2.2,3.5])
    spl_m0 = np.array([11.18,11.57,12.38,12.76])
    spl_K0 = np.array([9.04,9.04,8.99,8.79])
    m0 = np.interp(z, spl_z, spl_m0)
    K0 = np.interp(z, spl_z, spl_K0)
    metalZ_M08 = -0.0864 * (m - m0)**2 + K0
    print('calc_metalZ_from_FMR_following_Maiolino2008: logM0 = %s, K0 = %s'%(m0, K0))
    # 
    return metalZ_M08


def calc_metalZ_from_FMR_following_Magnelli2012(M_star, z):
    # metalZ is 12+log(O/H)_{Denicolo2002}
    # 
    # Magnelli2012 doi:10.1051/0004-6361/201220074
    # Sect. 4, Eq. 5
    # The Denicoló et al. calibration system provides the best agreement, at high stellar masses, between all different metallicity calibrators (Kewley & Ellison 2008)
    # We note that metallicities estimated using the “fundamental metallicity relation” of Mannucci et al. (2010) extended to z ∼ 2 in Genzel et al. (2012) are consistent with those derived using Eq. (5).
    m = np.log10(M_star)
    if z < 1.5:
        a = -4.45
    else:
        a = -4.51
    metalZ_D02 = (2.18 * m - 0.0896 * m**2 + a)
    # 
    return metalZ_D02


def calc_metalZ_from_FMR_following_Genzel2015_Eq12a_with_dzliu_limit(M_star, z):
    # Genzel et al. 2015:
    #   Following Maiolino et al. (2008) we combined the mass–metallicity relations 
    #   at different redshifts presented by Erb et al. (2006), 
    #   Maiolino et al. (2008), Zahid et al. (2014), and Wuyts et al. (2014) 
    #   in the following fitting function: 
    #     12+log(O/H) = a - 0.087 * (lgMstar - b)**2, 
    #     a = 8.74, and 
    #     b = 10.4(0.05) + 4.46(0.3) * lg(1+z) - 1.78(0.4) * (lg(1+z))**2  -- Eq. (12a)
    # 
    # dzliu_limit:
    #     when np.log10(M_star) > b, set (np.log10(M_star) - b) to zero. 
    # 
    a = 8.74
    b = 10.4 + 4.46 * np.log10(1+z) - 1.78 * (np.log10(1+z))**2
    #metalZ_PP04 = a - 0.087 * (np.log10(M_star) - b)**2
    diff = np.log10(M_star) - b
    if np.isscalar(diff):
        if diff<0.0:
            diff = 0.0 
    else:
        diff[(diff<0.0)] = 0.0
    # 
    metalZ_PP04 = a - 0.087 * (diff)**2
    # 
    return metalZ_PP04


def calc_metalZ_from_FMR_following_Genzel2015_Eq12a(M_star, z):
    # Genzel et al. 2015:
    #   Following Maiolino et al. (2008) we combined the mass–metallicity relations 
    #   at different redshifts presented by Erb et al. (2006), 
    #   Maiolino et al. (2008), Zahid et al. (2014), and Wuyts et al. (2014) 
    #   in the following fitting function: 
    #     12+log(O/H) = a - 0.087 * (lgMstar - b)**2, 
    #     a = 8.74, and 
    #     b = 10.4(0.05) + 4.46(0.3) * lg(1+z) - 1.78(0.4) * (lg(1+z))**2  -- Eq. (12a)
    a = 8.74
    b = 10.4 + 4.46 * np.log10(1+z) - 1.78 * (np.log10(1+z))**2
    metalZ_PP04 = a - 0.087 * (np.log10(M_star) - b)**2
    # 
    return metalZ_PP04


def calc_metalZ_from_FMR_following_Genzel2015a(M_star, SFR, z):
    # Genzel et al. 2015:
    #   Following Maiolino et al. (2008) we combined the mass–metallicity relations 
    #   at different redshifts presented by Erb et al. (2006), 
    #   Maiolino et al. (2008), Zahid et al. (2014), and Wuyts et al. (2014) 
    #   in the following fitting function: 
    #     12+log(O/H) = a - 0.087 * (lgMstar - b)**2, 
    #     a = 8.74, and 
    #     b = 10.4(0.05) + 4.46(0.3) * lg(1+z) - 1.78(0.4) * (lg(1+z))**2  -- Eq. (12a)
    a = 8.74
    b = 10.4 + 4.46 * np.log10(1+z) - 1.78 * (np.log10(1+z))**2
    metalZ_PP04 = a - 0.087 * (np.log10(M_star) - b)**2
    # 
    return metalZ_PP04


def calc_metalZ_from_FMR_following_Genzel2015_Eq12b(M_star, SFR):
    # Genzel et al. 2015 method 2:
    #   Mannucci et al. (2010) presented evidence for a dependence of metallicity 
    #   on the SFR for z ∼ 0 SDSS galaxies at a given stellar mass (the fundamental 
    #   metallicity relation), yielding an alternative version of Equation (12a):
    #     dZM08 = 12+log(O/H)_{M08} - 8.69 = 0.21 + 0.39 * x - 0.2 * x**2 - 0.077 * x**3 + 0.064 * x**4
    #     with x = logMstar - 0.32 * logSFR - 10, and
    #     12+log(O/H)_{PP04} - 8.9 = -0.4408 + 0.7044 * dZM08 - 0.1602 * dZM08**2 - 0.4105 * dZM08**3 - 0.1898 * dZM08**4
    # 
    #x = np.log10(M_star) - 0.32 * np.log10(SFR) - 10
    #dZM08 = 0.21 + 0.39 * x - 0.2 * x**2 - 0.077 * x**3 + 0.064 * x**4
    # 
    dZM08 = calc_metalZ_from_FMR_following_Mannucci2010_Eq4(M_star, SFR) - 8.69
    # 
    #dZM08 = calc_metalZ_from_FMR_following_Mannucci2010_Eq2(M_star, SFR) - 8.69
    # 
    metalZ = 8.9 + (-0.4408 + 0.7044 * dZM08 - 0.1602 * dZM08**2 - 0.4105 * dZM08**3 - 0.1898 * dZM08**4)
    # 
    return metalZ


def calc_metalZ_from_FMR_following_Genzel2015b(M_star, SFR):
    # Genzel et al. 2015 method 2:
    #   Mannucci et al. (2010) presented evidence for a dependence of metallicity 
    #   on the SFR for z ∼ 0 SDSS galaxies at a given stellar mass (the fundamental 
    #   metallicity relation), yielding an alternative version of Equation (12a):
    #     dZM08 = 12+log(O/H)_{M08} - 8.69 = 0.21 + 0.39 * x - 0.2 * x**2 - 0.077 * x**3 + 0.064 * x**4
    #     with x = logMstar - 0.32 * logSFR - 10, and
    #     12+log(O/H)_{PP04} - 8.9 = -0.4408 + 0.7044 * dZM08 - 0.1602 * dZM08**2 - 0.4105 * dZM08**3 - 0.1898 * dZM08**4
    # 
    #x = np.log10(M_star) - 0.32 * np.log10(SFR) - 10
    #dZM08 = 0.21 + 0.39 * x - 0.2 * x**2 - 0.077 * x**3 + 0.064 * x**4
    # 
    dZM08 = calc_metalZ_from_FMR_following_Mannucci2010_Eq4(M_star, SFR) - 8.69
    # 
    #dZM08 = calc_metalZ_from_FMR_following_Mannucci2010_Eq2(M_star, SFR) - 8.69
    # 
    metalZ = 8.9 + (-0.4408 + 0.7044 * dZM08 - 0.1602 * dZM08**2 - 0.4105 * dZM08**3 - 0.1898 * dZM08**4)
    # 
    return metalZ


def calc_metalZ_from_FMR_following_Genzel2015ab_combined_by_dzliu(M_star, SFR, z):
    # dzliu combined: 
    #   Genzel et al. 2015 method 1 and Genzel et al. 2015 method 2
    # when lgMstar >= 10.0, we use calc_metalZ_from_FMR_following_Genzel2015b()
    # otherwise we use calc_metalZ_from_FMR_following_Genzel2015a() but scaled to make sure the prediction is smooth at lgMstar=10.0
    # 
    if np.isscalar(M_star):
        input_M_star = np.array([M_star])
    else:
        input_M_star = np.array(M_star)
    # 
    if np.isscalar(SFR):
        input_SFR = np.array([SFR])
    else:
        input_SFR = np.array(SFR)
    # 
    if np.isscalar(z):
        input_z = np.array([z])
    else:
        input_z = np.array(z)
    # 
    if len(input_SFR) == 1 and len(input_M_star) > 1:
        input_SFR = np.array([input_SFR[0]]*len(input_M_star))
    # 
    if len(input_z) == 1 and len(input_M_star) > 1:
        input_z = np.array([input_z[0]]*len(input_M_star))
    # 
    mask = (input_M_star>=1e10)
    output_metalZ = input_M_star*0.0
    output_metalZ[mask] = calc_metalZ_from_FMR_following_Genzel2015b(input_M_star[mask], input_SFR[mask])
    output_metalZ[~mask] = calc_metalZ_from_FMR_following_Genzel2015a(input_M_star[~mask], input_SFR[~mask], input_z[~mask])
    renorm_metalZ = calc_metalZ_from_FMR_following_Genzel2015b(1e10, input_SFR[~mask]) - \
                    calc_metalZ_from_FMR_following_Genzel2015a(1e10, input_SFR[~mask], input_z[~mask])
    # 
    output_metalZ[~mask] = output_metalZ[~mask] + renorm_metalZ
    # 
    # 20200524
    # for lgMstar > 11.9, z < 0.8, always set to 8.75
    mask = np.logical_and(M_star >= 10**11.9, z < 0.8)
    output_metalZ[mask] = 8.75
    #
    if np.isscalar(M_star) and np.isscalar(SFR) and np.isscalar(z):
        return float(output_metalZ[0])
    return output_metalZ


def calc_metalZ_from_FMR_following_Guo2016(M_star, SFR):
    # metalZ is 12+log(O/H)_{Maiolino2008}
    # 
    # Yicheng Guo et al. 2016 (doi:10.3847/0004-637X/822/2/103)
    # Abstract: 12 + log(O/H) = (5.83 ± 0.19)+(0.30 ± 0.02)*log(Mstar)
    # The [O III]/Hβ flux ratio is then converted to metallicity through the calibration of Maiolino et al. (2008, M08).
    m = np.log10(M_star)
    metalZ_M08 = 5.83 + 0.30 * m
    # 
    return metalZ_M08









def convert_metalZ_M08_to_metalZ_PP04(metalZ_M08):
    # PP04: Pettini 2004
    #       N2 \equiv lg(NII6583/Halpha)
    #       metalZ_PP04 \equiv 12+lg(O/H) = 9.37 + 2.03 * N2 + 1.26 * N2**2 + 0.32 * N2**3      (PP04 Eq.1)
    #       or metalZ_PP04 = 8.90 + 0.57 * N2                                                   (PP04 Eq.2, valid in the range -2.5 < N2 < -0.3)
    #       they mentoned that the first polynomial form is not significantly better than the second linear form. 
    # M08:  Maiolino 2008
    #       N2 = -0.7732 + 1.2357 * x - 0.2811 * x**2 - 0.7201 * x**3 - 0.3330 * x**4
    #       where x = 12+lg(O/H) - 8.69 = metalZ_M08 - 8.69
    # so, the conversion from metalZ_M08 to metalZ_PP04 is:
    #       metalZ_PP04 = 8.90 + 0.57 * N2
    #                   = 8.90 + 0.57 * (-0.7732 + 1.2357 * x - 0.2811 * x**2 - 0.7201 * x**3 - 0.3330 * x**4)
    # same as Genzel+2015ApJ...800...20G Eq.12b.
    # 
    x = metalZ_M08 - 8.69
    return 8.90 + 0.57 * (-0.7732 + 1.2357 * x - 0.2811 * x**2 - 0.7201 * x**3 - 0.3330 * x**4)


def convert_metalZ_M08_to_metalZ_PP04_N2_polynomial(metalZ_M08):
    # PP04: Pettini 2004
    #       N2 \equiv lg(NII6583/Halpha)
    #       metalZ_PP04 \equiv 12+lg(O/H) = 9.37 + 2.03 * N2 + 1.26 * N2**2 + 0.32 * N2**3      (PP04 Eq.1)
    #       or metalZ_PP04 = 8.90 + 0.57 * N2                                                   (PP04 Eq.2, valid in the range -2.5 < N2 < -0.3)
    #       they mentoned that the first polynomial form is not significantly better than the second linear form. 
    # M08:  Maiolino 2008
    #       N2 = -0.7732 + 1.2357 * x - 0.2811 * x**2 - 0.7201 * x**3 - 0.3330 * x**4
    #       where x = 12+lg(O/H) - 8.69 = metalZ_M08 - 8.69
    # so, the conversion from metalZ_M08 to metalZ_PP04 is:
    #       metalZ_PP04 = 9.37 + 2.03 * N2 + 1.26 * N2**2 + 0.32 * N2**3
    #                N2 = -0.7732 + 1.2357 * x - 0.2811 * x**2 - 0.7201 * x**3 - 0.3330 * x**4
    # same as Genzel+2015ApJ...800...20G Eq.12b.
    # 
    x = metalZ_M08 - 8.69
    N2 = -0.7732 + 1.2357 * x - 0.2811 * x**2 - 0.7201 * x**3 - 0.3330 * x**4
    metalZ_PP04 = 9.37 + 2.03 * N2 + 1.26 * N2**2 + 0.32 * N2**3
    return metalZ_PP04


def convert_metalZ_D02_to_metalZ_PP04(metalZ_D02):
    # D02:  Denicolo et al. (2002) Eq.(2)
    #       N2 \equiv lg(NII6584/Halpha)
    #       metalZ_D02 \equiv 12+lg(O/H) = 9.12 + 0.73 * N2
    #       they mentoned that the first polynomial form is not significantly better than the second linear form. 
    # so, the conversion from metalZ_D02 to metalZ_PP04 (N2)
    #       according to Kewley & Ellison (2008ApJ...681.1183K) Table 3 is:
    # 
    a = -444.7831
    b = 165.42600
    c = -20.202000
    d = 0.8249386
    x = metalZ_D02
    return (a + b * x + c * x**2 + d * x**3)


def convert_metalZ_KK04_to_metalZ_PP04(metalZ_KK04):
    # KK04: 
    #       N2 \equiv lg(NII6584/Halpha)
    #       
    # so, the conversion from metalZ_KK04 to metalZ_PP04 (N2) 
    #       according to Kewley & Ellison (2008ApJ...681.1183K) Table 3 is:
    # 
    a = 916.7484
    b = -309.54480
    c = 35.051680
    d = -1.3188000
    x = metalZ_KK04
    return (a + b * x + c * x**2 + d * x**3)


def convert_metalZ_KK04_to_metalZ_PP04_O3N2(metalZ_KK04):
    # KK04: 
    #       N2 \equiv lg(NII6584/Halpha)
    #       
    # so, the conversion from metalZ_KK04 to metalZ_PP04 (O3N2) 
    #       according to Kewley & Ellison (2008ApJ...681.1183K) Table 3 is:
    # 
    a = 631.2766
    b = -210.02090
    c = 23.483050
    d = -0.8704286
    x = metalZ_KK04
    return (a + b * x + c * x**2 + d * x**3)


def convert_metalZ_KD02_to_metalZ_PP04(metalZ_KD02):
    # KD02: 
    #       Kewley & Dopita (2002)
    #       
    # so, the conversion from metalZ_KD02 to metalZ_PP04 (N2) 
    #       according to Kewley & Ellison (2008ApJ...681.1183K) Table 3 is:
    # 
    a = 569.4927
    b = -192.51820
    c = 21.918360
    d = -0.8278840
    x = metalZ_KD02
    return (a + b * x + c * x**2 + d * x**3)


def convert_metalZ_KD02_to_metalZ_PP04_O3N2(metalZ_KD02):
    # KD02: 
    #       Kewley & Dopita (2002)
    #       
    # so, the conversion from metalZ_KD02 to metalZ_PP04 (O3N2) 
    #       according to Kewley & Ellison (2008ApJ...681.1183K) Table 3 is:
    # 
    a = 664.8453
    b = -225.75330
    c = 25.768880
    d = -0.9761368
    x = metalZ_KD02
    return (a + b * x + c * x**2 + d * x**3)


def convert_metalZ_PP04_N2_to_metalZ_PP04_O3N2(metalZ_PP04_N2):
    # PP04 (N2): 
    #       Pettini 2004
    #       N2 \equiv lg(NII6583/Halpha)
    #       metalZ_PP04_N2_linear = 8.90 + 0.57 * N2                                        (PP04 Eq.2, valid in the range -2.5 < N2 < -0.3)
    #       metalZ_PP04_N2_polynomial = 9.37 + 2.03 * N2 + 1.26 * N2**2 + 0.32 * N2**3      (PP04 Eq.1)
    #       
    # so, the conversion from metalZ_PP04_N2 to metalZ_PP04 (O3N2) 
    #       according to Kewley & Ellison (2008ApJ...681.1183K) Table 3 is:
    # 
    a = -8.0069
    b = 2.74353
    c = -0.093680
    d = 0.0
    x = metalZ_PP04_N2
    return (a + b * x + c * x**2 + d * x**3)


def convert_metalZ_PP04_N2_linear_to_metalZ_PP04_N2_polynomial(metalZ_PP04_N2_linear):
    # PP04 (N2): 
    #       Pettini 2004
    #       N2 \equiv lg(NII6583/Halpha)
    #       metalZ_PP04_N2_linear = 8.90 + 0.57 * N2                                        (PP04 Eq.2, valid in the range -2.5 < N2 < -0.3)
    #       metalZ_PP04_N2_polynomial = 9.37 + 2.03 * N2 + 1.26 * N2**2 + 0.32 * N2**3      (PP04 Eq.1)
    #       
    # so, the conversion from metalZ_PP04_N2_linear to metalZ_PP04_N2_polynomial:
    # 
    N2 = (metalZ_PP04_N2_linear - 8.90) / 0.57
    metalZ_PP04_N2_polynomial = 9.37 + 2.03 * N2 + 1.26 * N2**2 + 0.32 * N2**3
    return metalZ_PP04_N2_polynomial



















# 
# dzliu selection of the best function
# 
def calc_metalZ_from_FMR_with_dzliu_selection(M_star, SFR, z):
    # metalZ is 12+log(O/H)_{PP04}
    # use Genzel+2015 Eq.12a (variable: z) when it is lower than Mannucci+2010 Eq.4 (SFR)
    # but at the massive end, if Genzel+2015 Eq.12a decreases with increasing stellar mass, we set it to not decrease. 
    input_M_star = np.log10(M_star)
    if np.isscalar(M_star):
        input_M_star = np.array([M_star])
    else:
        input_M_star = np.array(M_star)
    if np.isscalar(SFR):
        input_SFR = np.array([SFR])
    else:
        input_SFR = np.array(SFR)
    if np.isscalar(z):
        input_z = np.array([z])
    else:
        input_z = np.array(z)
    # 
    if len(input_z) == 1 and len(input_M_star) > 1:
        input_z = np.array([input_z[0]]*len(input_M_star))
        #print('calc_metalZ_from_FMR_with_dzliu_selection() Replicating input_z')
    if len(input_SFR) == 1 and len(input_M_star) > 1:
        input_SFR = np.array([input_SFR[0]]*len(input_M_star))
        #print('calc_metalZ_from_FMR_with_dzliu_selection() Replicating input_SFR')
    # 
    metalZ_G15a = calc_metalZ_from_FMR_following_Genzel2015_Eq12a(input_M_star, input_z)
    #metalZ_G15b = calc_metalZ_from_FMR_following_Genzel2015_Eq12b(input_M_star, input_SFR) # equals convert_metalZ_M08_to_metalZ_PP04_N2_linear(calc_metalZ_from_FMR_following_Mannucci2010_Eq4(input_M_star, input_SFR))
    metalZ_M10Eq4 = convert_metalZ_M08_to_metalZ_PP04_N2_polynomial(calc_metalZ_from_FMR_following_Mannucci2010_Eq4(input_M_star, input_SFR)) # apply convert_metalZ_M08_to_metalZ_PP04_N2_polynomial() conversion is better because it will turn down a little bit the metalZ at the low mass end!
    # 
    # choose min of (metalZ_G15a, metalZ_M10Eq4)
    mask1 = (metalZ_G15a < metalZ_M10Eq4)
    metalZ_out = metalZ_M10Eq4
    metalZ_out[mask1] = metalZ_G15a[mask1]
    #for i in range(1,len(metalZ_out)):
    #    if metalZ_out[i] < metalZ_out[i-1]:
    #        metalZ_out[i] = metalZ_out[i-1]
    # 
    # ref Mstar, above which we do not use G15 but use G15(ref Mstar)
    ref_M_star = 10**(10.4 + 4.46 * np.log10(1+input_z) - 1.78 * (np.log10(1+input_z))**2) # characteristic stellar mass deduced from Genzel2015_Eq12a
    ref_metalZ_G15a = calc_metalZ_from_FMR_following_Genzel2015_Eq12a(ref_M_star, input_z)
    #ref_metalZ_G15b = calc_metalZ_from_FMR_following_Genzel2015_Eq12b(ref_M_star, input_SFR) # equals convert_metalZ_M08_to_metalZ_PP04_N2_linear(calc_metalZ_from_FMR_following_Mannucci2010_Eq4(ref_M_star, input_SFR))
    #ref_metalZ_M10Eq4 = convert_metalZ_M08_to_metalZ_PP04_N2_polynomial(calc_metalZ_from_FMR_following_Mannucci2010_Eq4(ref_M_star, input_SFR)) # apply convert_metalZ_M08_to_metalZ_PP04_N2_polynomial() conversion is better because it will turn down a little bit the metalZ at the low mass end!
    # 
    # choose min of (metalZ_G15a, metalZ_M10Eq4) when (M_star < ref_M_star)
    #metalZ_out = metalZ_M10Eq4
    #mask1 = np.logical_and(metalZ_G15a < metalZ_M10Eq4, input_M_star < ref_M_star)
    #metalZ_out[mask1] = metalZ_G15a[mask1]
    # 
    # keeps constant at massive end
    #mask2 = (ref_metalZ_G15a < ref_metalZ_M10Eq4)
    #ref_metalZ_out = ref_metalZ_M10Eq4
    #ref_metalZ_out[mask2] = ref_metalZ_G15a[mask2]
    #mask3 = (input_M_star >= ref_M_star)
    #metalZ_out[mask3] = ref_metalZ_out[mask3]
    # 
    # keeps constant at massive end but use G15a
    mask3 = (input_M_star >= ref_M_star)
    metalZ_out[mask3] = ref_metalZ_G15a[mask3]
    # 
    return metalZ_out




















#!/usr/bin/env python
# 

from __future__ import print_function

import os, sys, re, json, time, astropy
import numpy as np
from astropy.table import Table, Column, hstack
from copy import copy
from numpy import log10, power as pow

from astropy.cosmology import FlatLambdaCDM
cosmo = FlatLambdaCDM(H0=73, Om0=0.27, Tcmb0=2.725)

if sys.version_info.major >= 3:
    long = int
else:
    pass



# 
# def 
# 
def convert_age_to_z(cosmoAge):
    # reversedly compute z from age
    if type(cosmoAge) is list: cosmoAge = np.array(cosmoAge)
    spl_opz_log10 = np.linspace(np.log10(1.0+0.0), np.log10(1.0+9999.0), num=1000, endpoint=True) # for converting cosmoAge to opz (1+z)
    spl_z = 10**spl_opz_log10-1.0
    spl_cosmoAge = cosmo.age(spl_z).value
    spl_cosmoAge_log10 = np.log10(spl_cosmoAge)
    spl_cosmoAge_log10 = spl_cosmoAge_log10[::-1]
    spl_opz_log10 = spl_opz_log10[::-1]
    tmp_opz_log10 = np.interp(np.log10(cosmoAge), spl_cosmoAge_log10, spl_opz_log10)
    z = 10**tmp_opz_log10-1.0
    return z


# CSFRD Madau & Dickinson 2014
def calc_cosmic_star_formation_rate_density_MadauDickinson2014(z):
    # Madau & Dickinson (2014)
    # converted to Chabrier IMF from Salpeter IMF by a factor of 1.64
    rho_SFR = 0.015 * (1+z)**2.7 / (1.0 + ((1+z)/2.9)**5.6) / 1.64
    return rho_SFR

def calc_CSFRD_Madau2014(z):
    return calc_cosmic_star_formation_rate_density_MadauDickinson2014(z)


# CSFRD Gladders et al. 2013, Abramson et al. 2016
def calc_cosmic_star_formation_rate_density_Abramson2016(z):
    # Gladders et al. 2013, Abramson et al. 2016
    A0 = 0.96 # see Abramson et al. 2016 Fig. 1 caption
    T0 = 1.64 # see Abramson et al. 2016 Fig. 1 caption
    tau = 0.66 # see Abramson et al. 2016 Fig. 1 caption
    t = cosmo.age(z).value
    rho_SFR = A0 / (t * np.sqrt(2.0*np.pi*(tau**2))) * np.exp((-(np.log(t)-T0)**2) / (2.0*(tau**2)))
    #rho_SFR = 0.96 * rho_SFR # http://arxiv.org/pdf/1604.00016.pdf Fig.1 Caption AUni=0.96
    rho_SFR = rho_SFR / 1.64 # Abramson et al. 2016 used Salpeter IMF while we convert to Chabrier IMF
    return rho_SFR

def calc_cosmic_star_formation_rate_density_Gladders2013(z):
    return calc_cosmic_star_formation_rate_density_Abramson2016(z)
    
def calc_CSFRD_Gladders2013(z):
    return calc_cosmic_star_formation_rate_density_Abramson2016(z)

def calc_CSFRD_Abramson2016(z):
    return calc_cosmic_star_formation_rate_density_Abramson2016(z)


# CSFRD Liudz 2018
def calc_cosmic_star_formation_rate_density_Liu2018(z, shape = 'double-powerlaw'):
    # D. Liu et al. 2018 (bibcode:2018ApJ...853..172L)
    # Chabrier IMF
    # shape can be 'double-powerlaw' or 'log-normal'
    if shape.startswith('double'):
        #rho_SFR = 0.00587 * (1+z)**3.0 / (1.0 + ((1+z)/2.9)**5.6)
        rho_SFR = 0.00962 * (1+z)**3.0 / (1.0 + ((1+z)/2.9)**5.6) # errtam: 0.015*10**(-0.1927) = 0.00962
        # Note that Madau & Dickinson parameters are: 0.015/1.64 * (1+z)**2.7 / (1.0 + ((1+z)/2.9)**5.6)
    elif shape.startswith('log'):
        A0 = 0.575
        T0 = 1.50
        tau = 0.66
        t = cosmo.age(z).value
        rho_SFR = A0 / (t * np.sqrt(2.0*np.pi*(tau**2))) * np.exp((-(np.log(t)-T0)**2) / (2.0*(tau**2)))
    else:
        raise ValueError('Error! The input shape %s is not allowed by the called function calc_cosmic_star_formation_rate_density_Liu2018()!'%(shape))
    return rho_SFR

def calc_CSFRD_Liu2018(z, shape = 'double-powerlaw'):
    return calc_cosmic_star_formation_rate_density_Liu2018(z, shape = shape)









#!/usr/bin/env python
# 
# Usage:
#    from calc_gas_mass_from_line import calc_gas_mass_from_line
# 
# 

from __future__ import print_function

import os, sys, re, json, time, astropy
import numpy as np
from astropy import units as u
from astropy.table import Table, Column, hstack
from copy import copy

from astropy.cosmology import FlatLambdaCDM
cosmo = FlatLambdaCDM(H0=70, Om0=0.27, Tcmb0=2.725)

if sys.version_info.major >= 3:
    long = int
else:
    pass

from calc_metal_Z import calc_metalZ_from_FMR_following_Genzel2015a
from calc_alpha_CO import calc_alphaCO_from_metalZ_following_Genzel2015a


# ============================================================================
# MOLECULAR GAS MASS CALCULATIONS
# ============================================================================

# 转换因子
ALPHA_CII = 30.0  # L_sun/M_sun ([CII] 光度 - 气体质量转换因子，文献范围 30-100)
ALPHA_CO_MW = 4.3  # M_sun/(K km/s pc²) (银河系)
ALPHA_CO_SB = 0.8  # M_sun/(K km/s pc²) (星暴/高红移)

def calc_gas_mass_from_line_flux(
        z, 
        lgMstar, 
        freq, 
        flux, 
        line_name, 
        cosmo = cosmo,
        alpha_CO = None,
        alpha_CII = None, 
        line_excitation_ratio = None,
        verbose = False,
    ):
    """Calculate (molecular) gas mass

    Return: (log_M_gas, log_L_line, line_unit, line_name)
      - log_M_gas: (molecular) gas mass (M_sun)
      - log_L_line: line luminosity ('L_sun' or 'K km s-1 pc2')
      - line_unit: line luminosity unit ('L_sun' or 'K km s-1 pc2')
      - line_type: line type ('CII' or 'CO')
    """
    lgMgas = float(np.nan)
    lgLline = float(np.nan)
    line_unit = 'unknown'
    line_type = 'unknown'
    
    if z is None or str(z) == '' or np.isnan(float(z)) or float(z) <= 0 or np.isnan(float(flux)) or float(flux) <= 0:
        return lgMgas, lgLline, line_unit, line_type
    
    # 光度距离
    D_L = cosmo.luminosity_distance(z).to(u.Mpc).value  # Mpc
    
    if line_name.startswith('[CII]') or line_name.startswith('CII'):
        # [CII] 光度 (L_sun)
        # L = 4π × D_L² × F × hν / (1+z)
        # F (Jy km/s) → F (erg/s/cm²)
        # 1 Jy = 10^{-23} erg/s/cm²/Hz
        # F_nu = F / Δv (假设线宽)
        # 简化：L_[CII] (L_sun) ≈ 1.04×10^-3 × (D_L/Mpc)² × (F/Jy km/s) × (ν_rest/GHz) / (1+z)
        # 简化：L_[CII] (L_sun) ≈ 1.04×10^-3 × (D_L/Mpc)² × (F/Jy km/s) × (ν_obs/GHz)
        
        L_CII_Lsun = 1.04e-3 * (D_L)**2 * flux * freq
        
        # 气体质量
        if alpha_CII is None:
            alpha_CII = ALPHA_CII
        M_gas = L_CII_Lsun / alpha_CII  # M_sun
        
        lgMgas = np.log10(M_gas)
        lgLline = np.log10(L_CII_Lsun)
        line_unit = 'L_sun'
        line_type = 'CII'
    
    elif line_name.startswith('CO'):
        # CO 光度 (K km/s pc²)
        # L'_CO = 3.25×10^7 × S_CO × ν_obs^{-2} × D_L² × (1+z)^{-3}
        # S_CO in Jy km/s, ν_obs in GHz, D_L in Mpc
        
        L_prime_CO = 3.25e7 * flux * (freq)**(-2) * (D_L)**2 * (1+z)**(-3)  # K km/s pc²

        # 激发改正
        r_J1_map = {
            'CO(1-0)': 1.0,
            'CO(2-1)': 0.8,
            'CO(3-2)': 0.5,
            'CO(4-3)': 0.4,
            'CO(5-4)': 0.3,
            'CO(6-5)': 0.25,
            'CO(7-6)': 0.2,
            'CO(8-7)': 0.15,
        }
        if line_excitation_ratio is None:
            r_J1 = r_J1_map.get(line_name, 0.5)
        else:
            if line_excitation_ratio > 1.0:
                raise ValueError(f'line_excitation_ratio ({line_excitation_ratio}) must be <= 1.0')
            r_J1 = line_excitation_ratio
        L_prime_CO10 = L_prime_CO / r_J1
        
        # 气体质量 (使用高红移质量依赖的 α_CO)
        if alpha_CO is None: 
            if lgMstar is None:
                alpha_CO = ALPHA_CO_SB
            else:
                SFR = None # not needed
                metalZ = calc_metalZ_from_FMR_following_Genzel2015a(10**lgMstar, SFR, z)
                alpha_CO = calc_alphaCO_from_metalZ_following_Genzel2015a(metalZ)
                if verbose:
                    print(f'metalZ: {metalZ:.2f}, Z/Zsun: {10**(metalZ-8.69):.2f}, alpha_CO: {alpha_CO:.2f}')
        M_gas = alpha_CO * L_prime_CO10  # M_sun
        if verbose:
            print(f'alpha_CO: {alpha_CO:.2f}, L_prime_CO: {L_prime_CO:.3e}, L_prime_CO10: {L_prime_CO10:.3e}, M_gas: {M_gas:.3e}')
        
        lgMgas = float(np.log10(M_gas))
        lgLline = float(np.log10(L_prime_CO)) # noting that the return is L_prime_CO not L_prime_CO10
        line_unit = 'K km s-1 pc2'
        line_type = 'CO'
    
    return lgMgas, lgLline, line_unit, line_type





import sys
import os
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', 'a3cosmos_gas_evolution', 'Common_Python_Code'))

import numpy as np
import pytest
from calc_cosmic_star_formation_rate_density import (
    convert_age_to_z,
    calc_cosmic_star_formation_rate_density_MadauDickinson2014,
    calc_CSFRD_Madau2014,
    calc_cosmic_star_formation_rate_density_Abramson2016,
    calc_cosmic_star_formation_rate_density_Gladders2013,
    calc_CSFRD_Gladders2013,
    calc_CSFRD_Abramson2016,
    calc_cosmic_star_formation_rate_density_Liu2018,
    calc_CSFRD_Liu2018,
)


class TestCalcCosmicSFRD:

    def test_convert_age_to_z_scalar(self):
        cosmoAge = 5.0
        result = convert_age_to_z(cosmoAge)
        assert result > 0

    def test_convert_age_to_z_array(self):
        cosmoAge = np.array([2.0, 5.0, 10.0])
        result = convert_age_to_z(cosmoAge)
        assert len(result) == 3
        assert np.all(result > 0)

    def test_madau_dickinson2014_scalar(self):
        result = calc_cosmic_star_formation_rate_density_MadauDickinson2014(1.0)
        assert result > 0

    def test_madau_dickinson2014_array(self):
        z = np.array([0.5, 1.0, 2.0, 3.0])
        result = calc_cosmic_star_formation_rate_density_MadauDickinson2014(z)
        assert len(result) == 4
        assert np.all(result > 0)

    def test_madau_dickinson2014_z0(self):
        result = calc_cosmic_star_formation_rate_density_MadauDickinson2014(0.0)
        assert result > 0

    def test_madau_dickinson2014_high_z(self):
        result = calc_cosmic_star_formation_rate_density_MadauDickinson2014(5.0)
        assert result > 0

    def test_CSFRD_Madau2014(self):
        result = calc_CSFRD_Madau2014(1.0)
        assert result > 0

    def test_abramson2016_scalar(self):
        result = calc_cosmic_star_formation_rate_density_Abramson2016(1.0)
        assert result > 0

    def test_abramson2016_z0(self):
        result = calc_cosmic_star_formation_rate_density_Abramson2016(0.0)
        assert result > 0

    def test_abramson2016_high_z(self):
        result = calc_cosmic_star_formation_rate_density_Abramson2016(3.0)
        assert result > 0

    def test_Gladders2013_alias(self):
        result_gladders = calc_cosmic_star_formation_rate_density_Gladders2013(1.0)
        result_abramson = calc_cosmic_star_formation_rate_density_Abramson2016(1.0)
        assert np.isclose(result_gladders, result_abramson)

    def test_CSFRD_Gladders2013_alias(self):
        result = calc_CSFRD_Gladders2013(1.0)
        assert result > 0

    def test_CSFRD_Abramson2016_alias(self):
        result = calc_CSFRD_Abramson2016(1.0)
        assert result > 0

    def test_liu2018_double_powerlaw(self):
        result = calc_cosmic_star_formation_rate_density_Liu2018(1.0, shape='double-powerlaw')
        assert result > 0

    def test_liu2018_log_normal(self):
        result = calc_cosmic_star_formation_rate_density_Liu2018(1.0, shape='log-normal')
        assert result > 0

    def test_liu2018_invalid_shape(self):
        with pytest.raises(ValueError):
            calc_cosmic_star_formation_rate_density_Liu2018(1.0, shape='invalid')

    def test_CSFRD_Liu2018_alias(self):
        result = calc_CSFRD_Liu2018(1.0)
        assert result > 0

    def test_CSFRD_Liu2018_with_shape(self):
        result = calc_CSFRD_Liu2018(1.0, shape='log-normal')
        assert result > 0

    def test_rhoSFR_positive_at_all_z(self):
        z_values = np.array([0.0, 0.5, 1.0, 2.0, 3.0, 5.0])
        for z in z_values:
            result = calc_cosmic_star_formation_rate_density_MadauDickinson2014(z)
            assert result >= 0

    def test_rhoSFR_has_peak_at_intermediate_z(self):
        z = np.array([0.0, 0.5, 1.0, 1.5, 2.0, 3.0, 5.0, 8.0])
        rho = calc_cosmic_star_formation_rate_density_MadauDickinson2014(z)
        peak_idx = np.argmax(rho)
        assert 0 < peak_idx < len(z) - 1



import sys
import os
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', 'a3cosmos_gas_evolution', 'Common_Python_Code'))

import numpy as np
import pytest
from calc_delta_GD import (
    calc_deltaGD_from_metalZ_following_Leroy2011,
    calc_deltaGD_from_metalZ_following_Magdis2012,
    calc_deltaGD_from_metalZ_following_RemyRuyer2014a,
    calc_deltaGD_from_metalZ_following_RemyRuyer2014b,
    calc_deltaGD_from_metalZ_following_Genzel2015,
)


class TestCalcDeltaGD:

    def test_leroy2011_scalar(self):
        result = calc_deltaGD_from_metalZ_following_Leroy2011(8.67)
        assert result > 0

    def test_leroy2011_array(self):
        metalZ = np.array([8.0, 8.5, 8.67, 9.0])
        result = calc_deltaGD_from_metalZ_following_Leroy2011(metalZ)
        assert len(result) == 4
        assert np.all(result > 0)

    def test_leroy2011_low_metallicity(self):
        result = calc_deltaGD_from_metalZ_following_Leroy2011(7.5)
        assert result > 0

    def test_leroy2011_high_metallicity(self):
        result = calc_deltaGD_from_metalZ_following_Leroy2011(9.5)
        assert result > 0

    def test_magdis2012_scalar(self):
        result = calc_deltaGD_from_metalZ_following_Magdis2012(8.67)
        assert result > 0

    def test_magdis2012_array(self):
        metalZ = np.array([8.0, 8.5, 8.67, 9.0])
        result = calc_deltaGD_from_metalZ_following_Magdis2012(metalZ)
        assert len(result) == 4
        assert np.all(result > 0)

    def test_remy_ruyer_2014a_scalar(self):
        result = calc_deltaGD_from_metalZ_following_RemyRuyer2014a(8.67)
        assert result > 0

    def test_remy_ruyer_2014a_below_knee(self):
        result = calc_deltaGD_from_metalZ_following_RemyRuyer2014a(7.5)
        assert result > 0

    def test_remy_ruyer_2014a_above_knee(self):
        result = calc_deltaGD_from_metalZ_following_RemyRuyer2014a(8.69)
        assert result > 0

    def test_remy_ruyer_2014a_array(self):
        metalZ = np.array([7.5, 8.0, 8.67, 9.0])
        result = calc_deltaGD_from_metalZ_following_RemyRuyer2014a(metalZ)
        assert len(result) == 4
        assert np.all(result > 0)

    def test_remy_ruyer_2014b_scalar(self):
        result = calc_deltaGD_from_metalZ_following_RemyRuyer2014b(8.67)
        assert result > 0

    def test_remy_ruyer_2014b_array(self):
        metalZ = np.array([7.5, 8.0, 8.67, 9.0])
        result = calc_deltaGD_from_metalZ_following_RemyRuyer2014b(metalZ)
        assert len(result) == 4
        assert np.all(result > 0)

    def test_genzel2015_scalar(self):
        result = calc_deltaGD_from_metalZ_following_Genzel2015(8.67)
        assert result > 0

    def test_genzel2015_same_as_leroy2011(self):
        metalZ = 8.67
        result_genzel = calc_deltaGD_from_metalZ_following_Genzel2015(metalZ)
        result_leroy = calc_deltaGD_from_metalZ_following_Leroy2011(metalZ)
        assert np.isclose(result_genzel, result_leroy)

    def test_deltaGD_decreases_with_metallicity(self):
        metalZ_low = 8.0
        metalZ_high = 9.0
        result_low = calc_deltaGD_from_metalZ_following_Leroy2011(metalZ_low)
        result_high = calc_deltaGD_from_metalZ_following_Leroy2011(metalZ_high)
        assert result_low > result_high



import sys
import os
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', 'a3cosmos_gas_evolution', 'Common_Python_Code'))

import numpy as np
import pytest
from calc_alpha_CO import (
    calc_alphaCO_from_metalZ_following_Wilson1995,
    calc_alphaCO_from_metalZ_following_Magdis2012,
    calc_alphaCO_from_metalZ_following_Genzel2012,
    calc_alphaCO_from_metalZ_following_Genzel2015a,
    calc_alphaCO_from_metalZ_following_Genzel2015b,
    calc_alphaCO_from_metalZ_following_Bolatto2013,
    calc_alphaCO_from_metalZ_following_Accurso2017,
    calc_alphaCO_from_metalZ_following_Bertemes2018,
    calc_alphaCO_from_metalZ_following_Tacconi2018,
    calc_alphaCO_from_metalZ_following_Boselli2014,
)


class TestCalcAlphaCO:

    def test_wilson1995_scalar(self):
        result = calc_alphaCO_from_metalZ_following_Wilson1995(8.67)
        assert np.isscalar(result)
        assert result > 0

    def test_wilson1995_array(self):
        metalZ = np.array([8.0, 8.5, 8.67, 9.0])
        result = calc_alphaCO_from_metalZ_following_Wilson1995(metalZ)
        assert len(result) == 4
        assert np.all(result > 0)

    def test_wilson1995_solar_metalZ(self):
        result = calc_alphaCO_from_metalZ_following_Wilson1995(8.67)
        assert result > 0

    def test_magdis2012_scalar(self):
        result = calc_alphaCO_from_metalZ_following_Magdis2012(8.67)
        assert np.isscalar(result) or result.shape == ()

    def test_magdis2012_array(self):
        metalZ = np.array([8.0, 8.5, 8.67, 9.0])
        result = calc_alphaCO_from_metalZ_following_Magdis2012(metalZ)
        assert len(result) == 4
        assert np.all(result > 0)

    def test_genzel2012_scalar(self):
        result = calc_alphaCO_from_metalZ_following_Genzel2012(8.67)
        assert np.isscalar(result) or result.shape == ()

    def test_genzel2012_array(self):
        metalZ = np.array([8.0, 8.5, 8.67, 9.0])
        result = calc_alphaCO_from_metalZ_following_Genzel2012(metalZ)
        assert len(result) == 4
        assert np.all(result > 0)

    def test_genzel2015a_default_params(self):
        result = calc_alphaCO_from_metalZ_following_Genzel2015a(8.67)
        assert result > 0

    def test_genzel2015a_explicit_params(self):
        result = calc_alphaCO_from_metalZ_following_Genzel2015a(8.67, metalZ_solar=8.67, alphaCO_MilkyWay=4.36)
        assert result > 0

    def test_genzel2015b_default_params(self):
        result = calc_alphaCO_from_metalZ_following_Genzel2015b(8.67)
        assert result > 0

    def test_genzel2015b_explicit_params(self):
        result = calc_alphaCO_from_metalZ_following_Genzel2015b(8.67, metalZ_solar=8.67, alphaCO_MilkyWay=4.36)
        assert result > 0

    def test_genzel2015_consistency(self):
        result_a = calc_alphaCO_from_metalZ_following_Genzel2015a(8.67)
        result_b = calc_alphaCO_from_metalZ_following_Genzel2015b(8.67)
        assert np.isscalar(result_a) or result_a.shape == ()
        assert np.isscalar(result_b) or result_b.shape == ()

    def test_bolatto2013_high_sigma(self):
        result = calc_alphaCO_from_metalZ_following_Bolatto2013(8.67, Sigma_total=150.0)
        assert result > 0

    def test_bolatto2013_low_sigma(self):
        result = calc_alphaCO_from_metalZ_following_Bolatto2013(8.67, Sigma_total=50.0)
        assert result > 0

    def test_bolatto2013_array_input(self):
        Sigma_total = np.array([50.0, 100.0, 150.0])
        result = calc_alphaCO_from_metalZ_following_Bolatto2013(8.67, Sigma_total)
        assert len(result) == 3
        assert np.all(result > 0)

    def test_accuro2017_scalar(self):
        result = calc_alphaCO_from_metalZ_following_Accurso2017(8.67, DeltaMS=0.0)
        assert result > 0

    def test_accuro2017_array_metalZ(self):
        metalZ = np.array([8.0, 8.5, 8.67, 9.0])
        result = calc_alphaCO_from_metalZ_following_Accurso2017(metalZ, DeltaMS=0.0)
        assert len(result) == 4
        assert np.all(result > 0)

    def test_bertemes2018_scalar(self):
        result = calc_alphaCO_from_metalZ_following_Bertemes2018(8.67)
        assert result > 0

    def test_tacconi2018_scalar(self):
        result = calc_alphaCO_from_metalZ_following_Tacconi2018(8.67)
        assert result > 0

    def test_boselli2014_lgL_Hband(self):
        result = calc_alphaCO_from_metalZ_following_Boselli2014(10.0)
        assert result > 0

    def test_boselli2014_array(self):
        lgL_Hband = np.array([9.0, 10.0, 11.0])
        result = calc_alphaCO_from_metalZ_following_Boselli2014(lgL_Hband)
        assert len(result) == 3
        assert np.all(result > 0)

    def test_alphaCO_decreases_with_metalZ(self):
        metalZ_low = 8.0
        metalZ_high = 9.5
        result_low = calc_alphaCO_from_metalZ_following_Wilson1995(metalZ_low)
        result_high = calc_alphaCO_from_metalZ_following_Wilson1995(metalZ_high)
        assert result_low > result_high
